use std::time::Instant;

use crate::{
    components::{monster::*, Monster, Player, Position, Renderable},
    constants::*,
    levels::Level,
    tile_data::{self, TileType},
    vga_color as vga,
};
use bracket_lib::{prelude::*, random::RandomNumberGenerator};
use specs::{Builder, Entity, World, WorldExt};

pub struct Map {
    tiles: Vec<TileType>,
    tile_content: Vec<Option<Entity>>,
    border_fg: RGB,
    border_bg: RGB,
    gem_color: RGB,
}

impl BaseMap for Map {}

impl Algorithm2D for Map {
    fn dimensions(&self) -> Point {
        Point::new(MAP_WIDTH, MAP_HEIGHT)
    }
}

impl Default for Map {
    fn default() -> Self {
        Self::new()
    }
}

impl Map {
    pub fn new() -> Self {
        let mut rng = RandomNumberGenerator::new();
        Self {
            tiles: vec![TileType::Floor; MAP_SIZE],
            tile_content: vec![None; MAP_SIZE],
            border_fg: RGB::named(vga::get_by_index(rng.range(8, 15))),
            border_bg: RGB::named(vga::get_by_index(rng.range(1, 7) as usize)),
            gem_color: RGB::named(vga::get_by_index(rng.range(1, 15))),
        }
    }

    pub fn from_level(level: Level) -> Self {
        match level {
            Level::Normal(_data) => todo!(),
            Level::Randomized(data) => {
                let mut rng = RandomNumberGenerator::new();
                let mut map = Self::new();
                for (tile, count) in data {
                    for _ in 0..count {
                        loop {
                            let point = Point {
                                x: rng.range(0, MAP_WIDTH as i32),
                                y: rng.range(0, MAP_HEIGHT as i32),
                            };
                            if map.get_tile_at(point) == TileType::Floor {
                                map.set_tile_at(point, tile);
                                break;
                            }
                        }
                    }
                }
                loop {
                    let point = Point {
                        x: rng.range(0, MAP_WIDTH as i32),
                        y: rng.range(0, MAP_HEIGHT as i32),
                    };
                    if map.get_tile_at(point) == TileType::Floor {
                        map.set_tile_at(point, TileType::Player);
                        break;
                    }
                }
                map
            }
            Level::End => todo!(),
        }
    }

    pub fn spawn_entities(&self, ecs: &mut World) {
        for (index, tile) in self.tiles.iter().enumerate() {
            let point = self.index_to_point2d(index);
            match tile {
                TileType::Player => {
                    let player_entity = ecs
                        .create_entity()
                        .with(Position {
                            x: point.x,
                            y: point.y,
                        })
                        .with(Renderable {
                            glyph: to_cp437('☻'),
                            fg: RGB::named(vga::YELLOW_BRIGHT),
                            bg: RGB::named(vga::BLACK),
                        })
                        .with(Player {
                            last_moved: Instant::now(),
                        })
                        .build();
                    ecs.insert(point);
                    ecs.insert(player_entity);
                }
                TileType::Slow => {
                    let mut rng = RandomNumberGenerator::new();
                    ecs.create_entity()
                        .with(Position {
                            x: point.x,
                            y: point.y,
                        })
                        .with(Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Slow))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Slow),
                            bg: RGB::named(vga::BLACK),
                        })
                        .with(Monster {
                            kind: MonsterKind::Slow,
                            ticks_until_move: ticks_for_kind(MonsterKind::Slow),
                        })
                        .build();
                }
                TileType::Medium => {
                    let mut rng = RandomNumberGenerator::new();
                    ecs.create_entity()
                        .with(Position {
                            x: point.x,
                            y: point.y,
                        })
                        .with(Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Medium))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Medium),
                            bg: RGB::named(vga::BLACK),
                        })
                        .with(Monster {
                            kind: MonsterKind::Medium,
                            ticks_until_move: ticks_for_kind(MonsterKind::Medium),
                        })
                        .build();
                }
                TileType::Fast => {
                    let mut rng = RandomNumberGenerator::new();
                    ecs.create_entity()
                        .with(Position {
                            x: point.x,
                            y: point.y,
                        })
                        .with(Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Fast))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Fast),
                            bg: RGB::named(vga::BLACK),
                        })
                        .with(Monster {
                            kind: MonsterKind::Fast,
                            ticks_until_move: ticks_for_kind(MonsterKind::Fast),
                        })
                        .build();
                }
                _ => {}
            }
        }
    }

    pub fn clear_all_tile_content(&mut self) {
        self.tile_content.iter_mut().for_each(|tile| {
            *tile = None;
        });
    }

    pub fn get_tile_content(&self, index: usize) -> Option<Entity> {
        self.tile_content[index]
    }

    pub fn set_tile_content(&mut self, index: usize, entity: Entity) {
        self.tile_content[index] = Some(entity);
    }

    pub fn get_tiles(&self) -> &Vec<TileType> {
        &self.tiles
    }

    pub fn draw(&self, ctx: &mut BTerm) {
        // Border
        ctx.fill_region(
            Rect {
                x1: MAP_X as i32 - 1,
                x2: MAP_WIDTH as i32 + MAP_X as i32 + 1,
                y1: MAP_Y as i32 - 1,
                y2: MAP_HEIGHT as i32 + MAP_Y as i32 + 1,
            },
            to_cp437('▓'),
            self.border_fg,
            self.border_bg,
        );

        for (i, tile) in self.tiles.iter().enumerate() {
            let point = self.index_to_point2d(i);
            let (x, y) = (point.x as usize, point.y as usize);

            let data = tile_data::tile_data(*tile);
            ctx.set(
                x + MAP_X,
                y + MAP_Y,
                if *tile == TileType::Gem {
                    self.gem_color
                } else {
                    RGB::named(data.color_fg)
                },
                data.color_bg,
                data.glyph,
            );
        }
    }

    pub fn get_tile_at(&self, point: Point) -> TileType {
        self.tiles[self.point2d_to_index(point)]
    }

    pub fn set_tile_at(&mut self, point: Point, tile: TileType) {
        let index = self.point2d_to_index(point);
        self.tiles[index] = tile;
    }

    pub fn is_solid(&self, point: Point) -> bool {
        matches!(self.get_tile_at(point), TileType::Wall | TileType::Block)
    }
}
