use crate::{
    components::{
        monster::{self, damage_for_kind, ticks_for_kind},
        Monster, Player, Position, Renderable,
    },
    resources::{Clock, Map, Stats},
    tile_data::TileType,
};
use bracket_lib::{prelude::*, random::RandomNumberGenerator};
use specs::prelude::*;

pub struct MonsterAiSystem {}

#[allow(clippy::type_complexity)]
impl<'a> System<'a> for MonsterAiSystem {
    type SystemData = (
        Entities<'a>,
        ReadExpect<'a, Clock>,
        ReadExpect<'a, Point>,
        ReadExpect<'a, Map>,
        WriteExpect<'a, Stats>,
        WriteStorage<'a, Monster>,
        WriteStorage<'a, Position>,
        WriteStorage<'a, Renderable>,
        ReadStorage<'a, Player>,
    );

    fn run(
        &mut self,
        (
            entities,
            clock,
            player_pos,
            map,
            mut stats,
            mut monsters,
            mut positions,
            mut renderables,
            players,
        ): Self::SystemData,
    ) {
        let mut data = (&entities, &mut monsters, &mut positions, &mut renderables)
            .join()
            .collect::<Vec<_>>();

        for (entity, monster, position, renderable) in &mut data {
            if clock.has_ticked {
                monster.ticks_until_move -= 1;
                if monster.ticks_until_move <= 0 {
                    // Change glyph
                    let mut rng = RandomNumberGenerator::new();
                    if let Some(glyph) =
                        rng.random_slice_entry(&monster::glyphs_for_kind(monster.kind))
                    {
                        renderable.glyph = *glyph;
                    }

                    // Move monster
                    let (mut delta_x, mut delta_y) = (0, 0);
                    if player_pos.x < position.x {
                        delta_x = -1;
                    }
                    if player_pos.x > position.x {
                        delta_x = 1;
                    }
                    if player_pos.y < position.y {
                        delta_y = -1;
                    }
                    if player_pos.y > position.y {
                        delta_y = 1;
                    }
                    let destination = Point {
                        x: position.x + delta_x,
                        y: position.y + delta_y,
                    };

                    if let Some(e) = map.get_tile_content(map.point2d_to_index(destination)) {
                        if let Some(_player) = players.get(e) {
                            // TODO: Sound
                            stats.take_gems(damage_for_kind(monster.kind));
                            let _ = entities.delete(*entity);
                        }
                    } else {
                        match map.get_tile_at(destination) {
                            TileType::Wall => {}
                            TileType::Block => {
                                // TODO: Sound
                                let _ = entities.delete(*entity);
                            }
                            _ => {
                                position.x = destination.x;
                                position.y = destination.y;
                            }
                        }
                    }

                    monster.ticks_until_move = ticks_for_kind(monster.kind);
                }
            }
        }
    }
}
